/*
** Module   : WebHandler.java
** Abstract : starts a web client and redirects to the web client's embedded server URI
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20131113 Created initial version
** 002 MAG 20131129 Add authentication page. Handle GET and POST requests.
** 003 MAG 20140206 Send URL as referrer address via configuration file.
** 004 CA  20140206 Changes related to refactoring the spawner infrastructure (allow P2J process
**                  clients to be started, beside web chui clients).
** 005 MAG 20140303 Add authorization token.
** 006 MAG 20140723 Send authorization token as a query parameter instead a cookie.
** 007 OM  20150108 Passed null map as additional environment to WebClientSpawner.spawn().
** 008 GES 20150127 Cleanup code formatting.
** 009 GES 20150311 Modified this to support both chui and gui web clients.
*/

package com.goldencode.p2j.main;

import java.io.*;
import java.util.logging.*;
import javax.servlet.*;
import javax.servlet.http.*;
import org.eclipse.jetty.http.*;
import org.eclipse.jetty.server.*;
import org.eclipse.jetty.server.handler.*;
import com.goldencode.p2j.ui.client.chui.driver.web.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.PlatformHelper;

/**
 * Starts a web client and redirects the requestor to the embedded server URI.
 * <p>
 * On GET method a web page containing an authentication form is returned to caller.
 * The user should enter the user name and the password used to authenticate on
 * the target client platform (Linux / Windows) and press the "Login" button to submit the
 * form (which will submit/POST the form).
 * <p>
 * On POST method using the user submitted credentials a new web client process is spawned
 * with the user's operating-system login environment. On web client URI notification the user's
 * browser is redirected to this URI. If no response from the spawned client occurs within a
 * specified amount of time, the authentication page is reloaded with an error message.
 */
public class WebHandler 
extends AbstractHandler 
{
   /** Authorization token query parameter name */
   public static final String PARAM_TOKEN = "token";

   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(WebHandler.class.getName());
   
   /** The root portion of the ChUI client request name (part of the URL). */
   private static final String TARGET_CHUI_ROOT = "/chui";
   
   /** The root portion of the GUI client request name (part of the URL). */
   private static final String TARGET_GUI_ROOT = "/gui";
   
   /** Error message to display, if any. */
   private String errorMessage = "";
   
   /** Flag that denotes client type (<code>true</code> for GUI, <code>false</code> for ChUI). */
   private boolean isGui = false;
   
   /**
    * Handle HTTP requests.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response
    *           The object or a wrapper of the response.
    */
   public void handle(String              target, 
                      Request             base, 
                      HttpServletRequest  request,
                      HttpServletResponse response) 
   throws IOException, 
          ServletException
   {
      if (!base.isHandled())
      {
         boolean chui = TARGET_CHUI_ROOT.equalsIgnoreCase(target);
         boolean gui  = TARGET_GUI_ROOT.equalsIgnoreCase(target);
         
         // only process for our two known targets 
         if (chui || gui)
         {
            this.isGui = gui;
            
            if ("GET".equalsIgnoreCase(base.getMethod()))
            {
               doGet(target, base, request, response);
            }
            else if ("POST".equalsIgnoreCase(base.getMethod()))
            {
               doPost(target, base, request, response);
            }
         }
      }
   }

   /**
    * Handle GET requests. Send a Login dialog page back to user.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response
    *           The object or a wrapper of the response.
    */
   public void doGet(String              target, 
                     Request             base, 
                     HttpServletRequest  request,
                     HttpServletResponse response) 
   throws IOException,
          ServletException
   {
      // produce the page contents
      response.setContentType(MimeTypes.Type.TEXT_HTML.asString());
      response.setStatus(HttpServletResponse.SC_OK);
      
      PrintWriter writer = response.getWriter();
      
      try
      {
         StringBuilder sb = new StringBuilder("/");
         String packageName = WebHandler.class.getPackage().getName();
         sb.append(packageName.replace('.', '/')).append("/web_client.html");
         
         InputStream is = WebHandler.class.getResourceAsStream(sb.toString());
         BufferedReader reader = new BufferedReader(new InputStreamReader(is));
         
         String str = null;
         
         while (null != (str = reader.readLine()))
         {
            writer.println(handleReplacements(str));
         }
         
         writer.flush();
         
         reader.close();
      }
      
      catch (IOException ioe)
      {
         LOG.logp(Level.SEVERE,
                  "WebHandler.doGet()",
                  "",
                  LogHelper.generate("IOException!"),
                  ioe);
      }
      
      finally
      {
         // mark the request as handled
         base.setHandled(true);
      }
   }
   
   /**
    * Handle POST request. Spawn web client and redirect the request to the embedded server.
    * 
    * @param    target
    *           The target of the request - either a URI or a name.
    * @param    base
    *           The base request.
    * @param    request
    *           The object or a wrapper of the request.
    * @param    response
    *           The object or a wrapper of the response.
    */
   public void doPost(String              target, 
                      Request             base, 
                      HttpServletRequest  request,
                      HttpServletResponse response) 
   throws IOException, 
          ServletException
   {
      // mark as handled
      base.setHandled(true);
      
      // get referrer URL
      String referrer = base.getHeader(HttpHeader.REFERER.asString());
      
      // get form parameters;
      String userName = base.getParameter("usr");
      String password = base.getParameter("psw");
      
      // spawn a new chui web client.
      WebClientSpawner webSpawner = new WebClientSpawner(isGui);
      ((WebClientBuilderParameters) webSpawner.cfg).setReferrer(referrer);
      webSpawner.spawn(userName, password, null);
      
      String remoteUri = webSpawner.getRemoteUri();
                                  
      if (remoteUri != null)
      {
         String type = isGui ? "GUI" : "ChUI";
         
         LOG.logp(Level.INFO, 
                  "WebHandler.handle", 
                  "",
                  LogHelper.generate("Redirect %s client to %s", type, remoteUri));
                  
         // set authorization token and redirect
         Object store = webSpawner.getServerData();
         
         if (store instanceof ServerKeyStore)
         {
            String token = ((ServerKeyStore)store).getAuthorizationToken();
            if (token != null)
            {
               remoteUri = remoteUri + "?" + PARAM_TOKEN + "=" + token;
            }
         }
         
         response.sendRedirect(remoteUri);
      }
      else
      {
         // build error message 
         errorMessage = htmlErrorMessage(webSpawner.getExitCode(), userName);
         
         // return a page having containing error message
         doGet(target, base, request, response);
         
         // reset error message
         errorMessage = "";
      }
   }
   
   /**
    * Process the given string and make any replacements of parameter
    * values as needed. The replacement placeholder follow the JEE 
    * expression language (EL) format for immediate evaluation. 
    * <p>
    * Example: ${expression} 
    *
    * @param    text
    *           The text line that may need modifications.
    *
    * @return   The new text if a modification occurred or the original text
    *           if no modification was needed.
    */
   protected String handleReplacements(String text)
   {
      if (text.contains("${error}"))
      {
         text = text.replace("${error}", errorMessage);
      }
      
      if (text.contains("${client_ui_type}"))
      {
         text = text.replace("${client_ui_type}", isGui ? "GUI" : "ChUI");
      }
      
      return text;
   }
   
   /**
    * Build a HTML string containing a message error based on the exit code.
    * 
    * @param    exitCode
    *           Exit code.
    * @param    userName
    *           Account user name.
    *           
    * @return   An error message as a HTML string.
    */
   private String htmlErrorMessage(int exitCode, String userName)
   {
      StringBuilder sb = new StringBuilder("<h1 style=\"color:red\">");

      if (PlatformHelper.isUnderWindowsFamily())
      {
         switch (exitCode)
         {                        
            case -2:
               sb.append("Invalid user name or password does not match.");
               break;
            case -3:
               sb.append("Cannot create user environment.");
               break;
            case -4:
               sb.append("Cannot get user profile.");
               break;
            case -5:
               sb.append("Cannot set working directory.");
               break;
            case -6:
               sb.append("Cannot create process for user ").append(userName);
               break;
            case -7:
               sb.append("Cannot load dynamic libraries.");
               break;
            default:
               sb.append("Cannot start the embedded server.");
               break;
         }
      }
      else
      {
         switch (exitCode)
         {
            case 1:
               sb.append("Invalid user name.");
               break;
            case 2:
               sb.append("Password does not match.");
               break;
            case 3:
               sb.append("User root not allowed.");
               break;
            case 4:
               sb.append("PAM authentication failed.");
               break;
            case -1:
               sb.append("Spawner has no root access permissions.");
               break;
            default:
               sb.append("Cannot start the embedded server.");
               break;
         }
      }
      
      sb.append("</h1>");
      return sb.toString();
   }
}
