/*
** Module   : ColorRgb.java
** Abstract : a simple color structure containing RGB values
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 MAG 20140829 Created initial version.
** 002 MAG 20140911 Added asString method.
** 003 GES 20150424 Added fromString() convenience method and setGuiRgb() method.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * A simple color structure containing RGB values. 
 */
public class ColorRgb
implements Externalizable
{
   /** Red value */
   private int red;
   
   /** Green value */
   private int green;
   
   /** Blue value */
   private int blue;
   
   /** Color name */
   private String name;
   
   /**
    * Constructor.
    * 
    * @param   red
    *          Red value.
    * @param   green
    *          Green value.
    * @param   blue
    *          Blue value.
    */
   public ColorRgb(int red, int green, int blue)
   {
      this.red = red;
      this.green = green;
      this.blue = blue;
   }

   /**
    * Constructor.
    * 
    * @param   red
    *          Red value.
    * @param   green
    *          Green value.
    * @param   blue
    *          Blue value.
    * @param   name
    *          Color name.
    */
   public ColorRgb(int red, int green, int blue, String name)
   {
      this(red, green, blue);
      this.name = name;
   }

   /**
    * Copy constructor.
    * 
    * @param   rgb
    *          ColorRgb copy.
    */
   public ColorRgb(ColorRgb rgb)
   {
      this(rgb.red, rgb.green, rgb.blue, rgb.name);
   }

   /**
    * Helper to take a Java compatible color definition and create a new instance of this
    * class.
    * 
    * @param    rgb
    *           Color in octal, hexadecimal (0xFFFFFF or #FFFFFF) or decimal format.  This uses
    *           <code>Integer.decode()</code> to create the integer that is then split into
    *           rgb values.
    * @param    def
    *           Value to copy if the string input is invalid.
    *
    * @return   The new instance.
    */
   public static ColorRgb fromString(String rgb, ColorRgb def)
   {
      ColorRgb result = new ColorRgb(def);
      
      try
      {
         int value = Integer.decode(rgb);
         
         result.setGuiRgb(value);
      }
      
      catch (NumberFormatException nfe)
      {
         // nothing to do, just return the previously created result
      }
      
      return result;
   }

   /**
    * Get red value.
    * 
    * @return  see above.
    */
   public int getRed()
   {
      return red;
   }

   /**
    * Set red value.
    * 
    * @param   red
    *          New red value
    */
   public void setRed(int red)
   {
      this.red = red;
   }

   /**
    * Get green value.
    * 
    * @return  see above.
    */
   public int getGreen()
   {
      return green;
   }

   /**
    * Set green value.
    * 
    * @param green
    *        New green value.
    */
   public void setGreen(int green)
   {
      this.green = green;
   }

   /**
    * Get blue value.
    * 
    * @return  see above.
    */
   public int getBlue()
   {
      return blue;
   }

   /**
    * Set blue value.
    * 
    * @param   blue
    *          New blue value.
    */
   public void setBlue(int blue)
   {
      this.blue = blue;
   }

   /**
    * Get a combination of red, green and blue value P4GL compliant.
    * 
    * @return  see above.
    */
   public int getRgb()
   {
      return this.red | (this.green << 8) | (this.blue << 16);
   }

   /**
    * Set color values form a RGB value using Progress 4GL parsing of the integer.
    * 
    * @param   rgb
    *          A combination of red, green and blue value.
    */
   public void setRgb(int rgb)
   {
      this.red   = rgb & 0xff;
      this.green = (rgb >>  8) & 0xff;
      this.blue  = (rgb >> 16) & 0xff;
   }

   /**
    * Get a combination of red, green and blue value Java compliant.
    * 
    * @return  see above.
    */
   public int getGuiRgb()
   {
      return this.blue | (this.green << 8) | (this.red << 16);
   }

   /**
    * Set color values form a RGB value using Java parsing of the integer.
    * 
    * @param   rgb
    *          A combination of red, green and blue value.
    */
   public void setGuiRgb(int rgb)
   {
      this.blue  = rgb & 0xff;
      this.green = (rgb >>  8) & 0xff;
      this.red   = (rgb >> 16) & 0xff;
   }

   /**
    * Reports if the given object is the same color as this instance. This does not just
    * compare references, but instead it compares the contained rgb values.
    *
    * @param    obj
    *           The object to compare.
    *
    * @return   <code>true</code> if the objects are equivalent, <code>false</code> if the given
    *           instance is <code>null</code>, is not an instance of this class or if its rgb
    *           value is not equivalent.
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;
      
      if (obj == null || getClass() != obj.getClass())
         return false;

      ColorRgb other = (ColorRgb) obj;

      return (red == other.red) && (green == other.green) && (blue == other.blue);
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */   
   public void readExternal(ObjectInput in) 
   throws IOException, 
          ClassNotFoundException
   {
      red   = in.readInt();
      green = in.readInt();
      blue  = in.readInt();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out) 
   throws IOException
   {
      out.writeInt(red);
      out.writeInt(green);
      out.writeInt(blue);      
   }
   
   /**
    * Return a RGB as string comma separated.
    * 
    * @return  see above.
    */
   public String asString()
   {
      return name == null ? red + "," + green  + "," + blue : name;
   }
   
   /**
    * Return a string with the RGB value in "0xRRGGBB" format where the red, green and
    * blue values are each encoded as 2 hexadecimal digits.
    *
    * @return   The value as a hexadecimal string.
    */
   public String asHexString()
   {
      return String.format("0x%02x%02x%02x", red, green, blue);
   }
   
   /**
    * Get color name.
    * 
    * @return  see above.
    */
   protected String getName()
   {
      return name;
   }
   
   /**
    * Set color name.
    * 
    * @param   name
    *          Color name.
    */
   protected void setName(String name)
   {
      this.name = name;
   }
}
