/*
** Module   : Cell.java
** Abstract : container which holds the details for rendering a single row/column in ChUI 
**
** Copyright (c) 2011-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 SIY 20110112 Created initial version.
** 002 GES 20150414 Moved empty cell definition from another class.
*/

package com.goldencode.p2j.ui.client;

import com.goldencode.p2j.ui.*;

/**
 * Single screen cell consisting of a character and a color.
 */
public class Cell
{
   /** Immutable empty cell (space character with normal color). */
   public final static Cell EMPTY_CELL = new Cell(' ', Color.NORMAL);
   
   /** Cell character. */ 
   private final char character;

   /** Cell color. */
   private final Color color;

   /**
    * Constructor.
    *
    * @param   character
    *          Cell character.
    * @param   color
    *          Cell color.
    */
   public Cell(char character, Color color)
   {
      this.character = character;
      this.color = color;
   }

   /**
    * Get cell character.
    *
    * @return cell character
    */
   public char character()
   {
      return character;
   }

   /**
    * Get cell color.
    *
    * @return cell color
    */
   public Color color()
   {
      return color;
   }

   /**
    * Check if cell represents real character and color.
    *
    * @param   cell
    *          Cell instance to check.
    *
    * @return  <code>true</code> if cell can be used to display character.
    */
   public static boolean isValid(Cell cell)
   {
      return cell != null && cell.color != null;
   }

   /**
    * @see java.lang.Object#hashCode()
    */
   @Override
   public int hashCode()
   {
      int result = 31 + character;
      return 31 * result + ((color == null) ? 0 : color.hashCode());
   }

   /**
    * @see java.lang.Object#equals(java.lang.Object)
    */
   @Override
   public boolean equals(Object obj)
   {
      if (this == obj)
         return true;

      if (obj == null || getClass() != obj.getClass())
         return false;

      Cell other = (Cell) obj;

      if (character != other.character)
         return false;

      if (color == null)
      {
         if (other.color != null)
            return false;
      }
      else if (!color.equals(other.color))
         return false;

      return true;
   }

   /**
    * Check if specified cell represents a space character.
    *
    * @param   cell
    *          Cell to check.
    *
    * @return  <code>true</code> if cell holds a space character.
    */
   public static boolean isSpace(Cell cell)
   {
      return (cell != null) ? cell.character() == ' ' : false;
   }
}
