/*
** Module   : PushMessagesWorker.java
** Abstract : a thread used to push messages to the remote end point
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ----------------------Description--------------------------
** 001 MAG 20140220 Created initial version.
** 002 GES 20150308 Moved to a more generic package location.
*/

package com.goldencode.p2j.ui.client.driver.web;

import java.io.*;
import java.nio.*;
import java.util.*;
import java.util.concurrent.*;

import org.eclipse.jetty.websocket.api.*;

/**
 * For performance reasons the sending of messages to the remote end point via web sockets
 * should be done asynchronously. A queue is used to store the messages. Messages are extracted
 * from the queue in a FIFO order and are then pushed down to the remote end point.
 * <p>
 * IMPORTANT: Do not use logging anywhere inside this class. When an authentication plugin
 * is running during session establishment with the P2J server using logging no messages are 
 * written into the log and the application becomes deadlocked.
 */
public class PushMessagesWorker 
extends Thread
{
   /** Time in milliseconds to wait for the thread exit. */
   private static final long WAIT_TO_DIE = 5000;

   /** Keyboard input queue. */
   private volatile Deque<Object> messages = new ConcurrentLinkedDeque<>();
   
   /** Remote session. */
   private volatile Session session;  

   /** Remote endpoint. */
   private volatile RemoteEndpoint remote;  
   
   /** Running flag. */
   private volatile boolean running = true;
   
   /**
    * Constructor.
    * 
    * @param    session
    *           Remote end point connection.
    */
   public PushMessagesWorker(Session session)
   {
      this.setName("pushworker");
      this.session = session;
      this.remote = session.getRemote();
      this.start();
   }
   
   /**
    * Add a new message at the end of the FIFO queue and notify the push thread.
    * 
    * @param    message
    *           A binary or a text message.
    */
   public void pushMessage(Object message)
   {
      if (running)
      {
         messages.offerLast(message);
         
         // notify
         synchronized (messages)
         {
            messages.notify();
         }
      }
   }
   
   /**
    * Get messages from queue and send. Wait if no messages on queue.  
    */
   @Override                                          
   public void run()
   {
      while (running)
      {
         // wait until messages are available in queue; this may also return with the
         // queue empty if the thread has been terminated
         waitForMessages();
         
         // send a message from queue (this will check if the running flag is cleared
         // so it is safe even if the queue is empty)
         sendMessage();         
      }
   }

   /**
    * Kill thread and wait to die.
    */
   public void stopPushing()
   {
      running = false;
      this.interrupt();

      // wait for thread to exit      
      try
      {
         this.join(PushMessagesWorker.WAIT_TO_DIE);
      }
      catch (InterruptedException e)
      {
         e.printStackTrace();
      }
   }
   
   /**
    * Block if the queue is empty. This will return when there are messages available in the
    * queue OR in the case where the thread has been interrupted (this should be because the
    * thread has been shutdown but it could also happen spuriously). For this reason, the
    * caller must place this in a loop and check the <code>running</code> flag.
    */
   private void waitForMessages()
   {
      synchronized (messages)
      {
         // are messages in the queue ?
         if (messages.isEmpty())
         {
            try
            {
               // no messages on the queue, wait
               messages.wait();
            }
            catch (InterruptedException e)
            {
               e.printStackTrace();
            }
         }
      }      
   }
   
   /**
    * Send a message.
    */
   private void sendMessage()
   {
      // test if interrupted 
      if (!isInterrupted() && running)
      {
         // get a message from queue, may be null if the queue is empty and there was
         // a spurious interruption in the waiting method
         Object message = messages.pollFirst();

         // send message
         if (session.isOpen())
         {
            try
            {
               // test message type
               if (message instanceof ByteBuffer)
               {
                  remote.sendBytes((ByteBuffer) message);
               }
               else if (message instanceof String)
               {
                  remote.sendString((String) message);
               }
            }
            catch (IOException e)
            {
               running = false;
               e.printStackTrace();
            }
         }    
         else
         {
            running = false;
         }
      }      
   }
}
