/*
** Module   : p2j.sockets.js
** Abstract : WebSockets communication management
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**         Golden Code Development Corporation
**                 CONFIDENTIAL
**
** -#- -I- --Date-- ------------------------------Description----------------------------------
** 001 MAG 20140110 First version.
** 002 MAG 20140204 Get attributes as a binary message. On quit command go back to login page. 
** 003 MAG 20140211 Implements COLOR.
** 004 MAG 20140214 Implements referrer redirection.
** 005 MAG 20140219 Performance improvements.
** 006 MAG 20140402 Implements interactive child process mode.
** 007 GES 20150324 Moved this class to a common resource base and made it generic (so that it
**                  can be used for both ChUI and GUI web clients.
*/

p2j.socket = (function()
{
   /** Socket object which will be returned from this initialization function. */
   var me = {}; 

   /** Referrer address. */
   var referrer;

   /** Web socket. */
   var ws;
   
   /** Connected flag. */
   var connected = false;

   /**
    * Send data.
    * 
    * @param {array} data
    */
   me.send = function(data)
   {
      if (connected) 
      {
         ws.send(data);
      }
      else
      {
         p2j.sound.beep();
      }
   };

   /**
    * Initialize module.
    * 
    * @param {object} cfg  configuration.
    */
   me.init = function(cfg)
   {      
      referrer = cfg.referrer;
      
      if ('WebSocket' in window)
      {
         ws = new WebSocket(cfg.socket.url);

         // on web socket open
         ws.onopen = function()
         {
            connected = true;
            ws.binaryType = 'arraybuffer';
         };

         // on web socket message
         ws.onmessage = function(evt)
         {
            var data = evt.data;
            
            if (data instanceof ArrayBuffer)
            {
               // binary messages
               var message = new Uint8Array(data);
               
               switch(message[0])
               {
                  case 0x80:
                     // clear screen 
                     p2j.screen.clear();                        
                     break;
                  case 0x81:
                     // draw screen
                     p2j.screen.drawRectangles(message);
                     break;
                  case 0x82:
                     // set cursor position
                     p2j.screen.setCursorPosition(message[1], message[2]);                        
                     break;
                  case 0x83:
                     // show cursor
                     p2j.screen.setCursorStatus(message[1]);                        
                     break;
                  case 0x84:
                     // message beep
                     p2j.sound.beep();                        
                     break;
                  case 0x85:
                     // quit
                     window.location.replace(referrer);                        
                     break;
                  case 0x86:
                     // switch mode p2j/vt100
                     p2j.keyboard.vt100 = (message[1] == 0) ? false : true;          
                     break;
               };               
            }
            else
            {
               // text messages
               var pay = JSON.parse(data);
               
               switch (pay.c)
               {
                  case 0:
                     // color palette
                     p2j.screen.palette = pay.p;                        
                     break;
               };
            }
         };

         // on web socket close
         ws.onclose = function()
         { 
            connected = false;
            p2j.screen.error('Connection closed by remote host.'); 
            window.setInterval(function() { window.location.replace(referrer); }, 5000);
         };
         
         // refresh
         window.onpagehide = function() 
         {
            if (connected)
            {
               ws.close();
            }
         };                  
      } 
      else 
      {
         p2j.screen.error('WebSockets are NOT supported by your Browser.');
      }      
   };
   
   return me;       
}());
