/*
** Module   : StatusLineGuiImpl.java
** Abstract : window status line implementation 
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140522 First version.
** 002 CA  20140711 Enhanced font support; refactoring related to MESSAGE-AREA widget.
** 003 CA  20140805 Added generic's to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Some refactoring to match new drawing techniques.
** 004 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 005 MAG 20140912 Implements GUI status line.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 MAG 20140917 Changed background color.
** 009 CA  20141021 Removed the repaint() API: all repainting needs to be done via the common
**                  event processing; do no draw the widget immediately.
** 010 MAG 20141016 Change text vertical position. Eliminate GuiOptions.BORDER_SIZE dependency.
** 011 CA  20150323 The GuiFontResolver needs to know if the expected font is for a fixed-width
**                  widget.
** 012 GES 20150425 Eliminated AWT dependencies.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * In GUI implementation the window status line is implemented as widget.
 */
@SuppressWarnings("unchecked")
public class StatusLineGuiImpl
extends StatusLine<GuiOutputManager>
{
   /** Constants used to draw right-bottom corner */
   private static final int G = 1;

   /** Screen driver */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Parent window of this widget */
   private WindowGuiImpl window;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Widget height in pixels */
   private int widgetHeight;

   /** Widget height in characters */
   private double widgetHeightChars;

   /**
    * Hide base class public constructor.
    * <p>
    * Create a widget representing the window's status line.
    * 
    * @param    id
    *           The widget's ID.
    */
   StatusLineGuiImpl(WidgetId id)
   {
      super(id);
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      // set default BGCOLOR and FGCOLOR
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT);
      // color used to draw bottom right corner
      gc.miscColor = gd.getSysColor(SysColor.COLOR_WINDOW);
      // margins
      nativeInsets = new NativeInsets(2, 2, 2, 2);
   }

   /**
    * Widget drawing method.
    */
   @Override
   public void draw()
   {
      if (!isVisible())
      {
         return;
      }
      
      // resolve colors
      gc.colors();
      
      // resolve font
      final FontCache cached = gf.font();

      // get status line coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
      final NativeRectangle clip = new NativeRectangle(nativeInsets.top, 
                                                       nativeInsets.left,
                                                       d.height - nativeInsets.bottom,
                                                       d.width - nativeInsets.right);
      final String text = getLabelText();
            
      gd.draw(p, null, new Runnable()
      {
         @Override
         public void run()
         {
            // draw status bar
            gd.setColor(gc.bgColor);
            // clear the area
            gd.fillRect(0, 0, d.width, d.height);
            gd.draw3DRect(0, 0, d.width, d.height, false);

            // draw status bar text
            gd.setColor(gc.fgColor);
            gd.setGuiFont(cached.font);
            gd.setFontStyle(cached.font.style);
            gd.setDrawingArea(clip);
            gd.drawString(text, nativeInsets.left, nativeInsets.top + cached.font.pointSize);            
            gd.setDrawingArea(null);

            // if window is resizable draw a small triangle on bottom right corner
            if (window.config().resize)
            {
               // corner constants
               int W1 = widgetHeight * 3 / 4;
               int W2 = W1 - 1;
               
               // draw status bar right-bottom corner            
               int[] xPoints = {d.width - W1, d.width,  d.width};
               int[] yPoints = {d.height,     d.height, d.height - W1};
               gd.setColor(gc.miscColor);
               gd.fillPolygon(xPoints, yPoints);            
   
               int[] xPoints2 = {d.width - W2, d.width +  G, d.width  + G};
               int[] yPoints2 = {d.height + G, d.height + G, d.height - W2};
               gd.setColor(gc.bgColor);
               gd.fillPolygon(xPoints2, yPoints2);
            }
            
            // force draw
            gd.flush();
         }
      });
   }

   /**
    * Attach this widget to the specified window.
    * 
    * @param    window
    *           The parent window.
    */
   @Override
   public void attach(Window<GuiOutputManager> window)
   {
      this.window = (WindowGuiImpl) window;
      init();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      return widgetHeightChars;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      return window.workspace.width();
   }
   
   /**
    * Status bar initialization.
    */
   private void init()
   {
      config().font = window.resolveStatusAreaFont();
      gf = new GuiFontResolver(window, gd, config(), false);
      FontCache cached = gf.font();
      widgetHeight = cached.fontHeight + nativeInsets.top + nativeInsets.bottom;      
      widgetHeightChars = screen().coordinates().heightFromNative(widgetHeight);
   }   
}
