/* 
** Module   : TextGuiImpl.java
** Abstract : static text implementation for GUI
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.
** 002 CA  20150323 Fixed calculations of widget dimensions in GUI.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150406 Fixed text height in non-3d mode.
** 005 GES 20150425 Eliminated AWT dependencies.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Text;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.util.*;

/**
 * GUI-specific static text widget ({@link Text}) implementation.
 */
public class TextGuiImpl
extends Text<GuiOutputManager>
{
   /** Font resolver - for container purposes. */
   private GuiFontResolver gfd;

   /** Font resolver - for display purposes. */
   private GuiFontResolver gfi;

   /** Screen driver */
   private GuiDriver gd;

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public TextGuiImpl(WidgetId id, TextConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, TextConfig cfg)
   {
      super.initialize(id, cfg);

      boolean fixedFontDim = !config.flagStatic && config.appDataPres.isFixedFont();
      boolean fixedFontDisp = !config.flagStatic       && 
                              !config().explicitViewAs && 
                              config.appDataPres.isFixedFont();

      gd = (GuiDriver) OutputManager.getDriver();

      gfd = new GuiFontResolver(Window.resolveWindow(this), gd, config(), fixedFontDim);
      gfi = new GuiFontResolver(Window.resolveWindow(this), gd, config(), fixedFontDisp);
      
      if (cfg.frameDefValue != null)
      {
         setValue(cfg.frameDefValue);
         
         // use only once
         cfg.frameDefValue = null;
      }

      updateSize();
   }

   /**
    * Draw this text.
    */
   @Override
   public void draw(Color color)
   {
      if (assigned && getState() == ScreenBuffer.UNINITIALIZED)
         setState(ScreenBuffer.UNCHANGED);
      
      Color dc = UiUtils.getDColor(this);

      selected = !Utils.equals(color, dc);
      
      final FontCache fc = gfi.font();
      final FontDetails<?> font = fc.font;

      final GuiColorResolver gc = new GuiColorResolver(gd, config, 
                                                       SysColor.COLOR_FILLIN_BG,
                                                       SysColor.COLOR_FILLIN_FG);
      gc.colors();

      NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      final String txt = getText(true);
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // clear the area
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, dim.width, dim.height);

            // draw full text
            gd.setGuiFont(font);
            gd.setColor(gc.fgColor);

            // needs to be centered vertically
            int txtHeight = nativeHeight(true);
            int diff = (dim.height - txtHeight) / 2;
            int ytxt = txtHeight + diff;

            if (config.flagStatic)
            {
               gd.drawStringScaled(txt, nativeWidth(true), nativeHeight(true), 0, ytxt);
            }
            else
            {
               gd.drawString(txt, 0, ytxt);
            }
         }
      });
   }

   /**
    * Main worker routine for the text accessor.
    * 
    * @param   isInternal
    *          If this flag is set to <code>true</code> then alternative variable value is used 
    *          when available.
    *          
    * @return  Text representation of the variable value according to all current settings, 
    *          flags, format, etc.
    */
   @Override
   public String getText(boolean isInternal)
   {
      String display = getScreenValue(isInternal);
      
      if (isInternal)
      {
         // obtain the initialization status
         boolean uninit = (getState() == ScreenBuffer.UNINITIALIZED);
         
         // If uninitialized and not for input and not date fill-in (in this case
         // separators should be displayed), display blanks instead.
         // TODO: probably it should be replaced by variable initialization
         //       in FrameHelper.clearComponents().
         if (uninit &&
             !(config.appFormat != null &&
               config.appFormat instanceof DateFormat &&
               isEnabled()))
         {
            display = "";
         }
      }
      
      return display;
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public TextConfig config()
   {
      return (TextConfig) super.config();
   }

   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      if (config.fixedWidth)
      {
         return config.widthPixels;
      }

      return nativeWidth(false);
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      if (config.fixedHeight)
      {
         return config.heightPixels;
      }

      return nativeHeight(false);
   }

   /**
    * Retrieve the widget's width in native units.
    * 
    * @param    display
    *           <code>true</code> if this is called during drawing.
    *           
    * @return   See above.
    */
   private int nativeWidth(boolean display)
   {
      Window<?> window = Window.resolveWindow(this);
      ScreenDriver<?> driver = OutputManager.getDriver();
      GuiFontResolver gf = (display ? gfi : gfd);

      if (gf == null || config == null || config.appDataPres == null)
         return 0;

      int font = gf.resolveFontNum();
      
      // literals case
      String txt = getText(true);
      boolean atField = isAtField();
      
      if (display || (txt.length() > 0 && !atField && config().flagStatic))
      {
         return FontManager.getTextWidth(txt, window, font, driver);
      }
      else if (config().explicitViewAs || config.flagStatic)
      {
         return FillInGuiImpl.nativeWidth(this, gd, gf, true);
      }
      else
      {
         int charNo = atField ? config.atFormatLength : config.appFormat.getScreenWidth();
         
         return FontManager.getFontWidth(window, font, driver) * charNo;
      }
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @param    display
    *           <code>true</code> if this is called during drawing.
    *           
    * @return   See above.
    */
   private int nativeHeight(boolean display)
   {
      Window<?> window = Window.resolveWindow(this);
      ScreenDriver<?> driver = OutputManager.getDriver();
      GuiFontResolver gf = (display ? gfi : gfd);
      
      if (gf == null || config == null || config.appDataPres == null)
         return 0;

      int font = gf.resolveFontNum();
      // for text-widget height, always use the DEFAULT-FONT
      if (font == FontManager.DEFAULT_FIXED_FONT)
      {
         font = FontManager.DEFAULT_FONT;
      }
      String txt = getText(true);
      boolean threeD = FontManager.isThreeD();

      if (display)
      {
         // literals case
         return FontManager.getTextHeight(txt, window, font, driver);
      }
      else
      {
         return config.flagStatic && !config().explicitViewAs
                   ? screen().coordinates().heightToNative(1) - (threeD ? 2 : 0)
                   : FontManager.getFontHeight(window, font, driver);
      }
   }
}
