/*
** Module   : AbstractGuiDriver.java
** Abstract : simple base class for GUI screen driver implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150401 Created initial version.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

// WARNING: these AWT/ImageIO dependencies are being intentionally used in this generic GUI code
//          because there is no better alternative for non-Swing drivers and this code works
//          fine in headless mode
import java.awt.image.BufferedImage;
import javax.imageio.*;
import javax.imageio.stream.*;

import java.io.*;
import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Simple base class for GUI-specific screen driver implementation features. ChUI-specific
 * features are not to be implemented here.  This class is intended to be specific to a given
 * UI-modality (GUI in this case). However, no driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 * <p>
 * The only exceptions to this no-Swing rule are for events and image loading.  We are using
 * the AWT event classes as simple containers to eliminate the need to create our own versions
 * of the same thing. For image loading, there is no better solution than to use ImageIO and
 * AWT. 
 */
public abstract class AbstractGuiDriver<F>
extends AbstractDriver<GuiPrimitives, GuiOutputManager>
implements GuiDriver<F, BufferedImage>
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(AbstractGuiDriver.class.getName());
   
   /** Currently selected window emulator. */
   protected EmulatedWindowState ews = null;
   
   /** The real location of the translated origin. */
   protected NativePoint translatedOrigin = new NativePoint(0, 0);

   /** Pending tooltip workers to be registered with mouse handler. */
   private Map<Widget, MouseWidgetAction> pendingTtWorkers = new HashMap<>();
   
   /** System colors cache. */
   private Map<Integer, ColorRgb> systemColorsCache = new HashMap<>();

   /** The currently used theme for GUI widgets. */
   private String theme = null;

   /**
    * Construct a new instance.
    *
    * @param    direct
    *           The output primitives.
    */
   public AbstractGuiDriver(GuiPrimitives direct)
   {
      super(new GuiWidgetFactory(), direct, false, "win3");
   }
   
   /**
    * Get the {@link OutputManager} instance associated with this driver.
    * 
    * @param    cfg
    *           The initial configuration for this driver.
    *
    * @return   A {@link GuiOutputManager} instance.
    */
   @Override
   public GuiOutputManager createOutputManager(BootstrapConfig cfg)
   {
      return new GuiOutputManager(this, cfg);
   }
   
   /**
    * Clear the entire screen, with the default window color.
    */
   @Override
   public void clear()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR);
      ps.x = 0;
      ps.y = 0;
      ps.width = ews.getWidth();
      ps.height = ews.getHeight();
      ps.color = null; // TODO: color???
      
      ews.offer(ps);
   }

   /**
    * Create a native image from the raw bytes and return a generic wrapper that can be used
    * for operations.
    *
    * @param    byts
    *           The image as persisted in the file system or a JAR file.
    *
    * @return   The generic wrapper representing the given image or <code>null</code> if the
    *           image cannot be created (e.g. if the bytes are not a valid image type).
    */
   @Override
   public ImageWrapper<BufferedImage> createImage(byte[] byts)
   {
      // the new image candidate
      BufferedImage        img  = null;
      ByteArrayInputStream bais = null;

      // try creating new image
      try
      {
         bais = new ByteArrayInputStream(byts);
         img  = ImageIO.read(bais);
      }
      
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not load image from byte array...", e);
         }
      }
      
      finally
      {
         if (img == null && bais != null)
         {
            try
            {
               bais.close();
            }
            
            catch (IOException ioe)
            {
               // ignore, don't care
            }
         }
      }
      
      return img == null ? null
                         : new ImageWrapper<BufferedImage>(img, img.getHeight(), img.getWidth());
   }
   
   /**
    * Create a native icon from the raw bytes and return a generic wrapper that can be used
    * for operations.
    *
    * @param    byts
    *           The image as persisted in the file system or a JAR file.
    * @param    small
    *           <code>true</code> if the small icon size is to be loaded.
    *
    * @return   The generic wrapper representing the given icon or <code>null</code> if the
    *           icon cannot be created (e.g. if the bytes are not a valid icon type).
    */
   @Override
   public ImageWrapper<BufferedImage> createIcon(byte[] byts, boolean small)
   {
      BufferedImage        img = null;
      ByteArrayInputStream bais   = null;
      ImageReader          reader = null;
      
      try
      {
         bais = new ByteArrayInputStream(byts);
         
         ImageInputStream iis = ImageIO.createImageInputStream(bais);
         
         // get the list of readers
         Iterator<ImageReader> readers = ImageIO.getImageReaders(iis);

         if (readers.hasNext())
         {
            // get the reader, we only ever use the first one
            reader = readers.next();

            // connect reader to input stream
            reader.setInput(iis);
            
            // get the icon numbers in a stream allowing internal scanning
            int numIcons = reader.getNumImages(true);
            
            // calculate the size of the icon we are searching for
            int targetSize = small ? WindowConfig.SMALL_ICON_SIZE : WindowConfig.ICON_SIZE;
            
            // check all icons
            for (int i = 0; i < numIcons; i++)
            {
               int iconSize = Math.max(reader.getWidth(i), reader.getHeight(i));
               
               if (iconSize == targetSize)
               {
                  // found the required image, read it and exit icon searching loop
                  img = reader.read(i);
                  break;
               }
            }
         }

         // have not found required icon size or not icon, use first available image 
         if (img == null)
         {
            if (bais != null)
            {
               bais.close();
            }
            
            bais = new ByteArrayInputStream(byts);
            
            // if this succeeds, there is no need to close the stream, but if this returns
            // null then the stream needs closing
            img  = ImageIO.read(bais);
         }
      }
      
      catch (IOException e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not set icon from array...", e);
         }

         img = null;
      }
      
      finally
      {
         if (reader != null)
         {
            reader.dispose();
         }
         
         if (bais != null)
         {
            try
            {
               bais.close();
            }
            
            catch (IOException ioe)
            {
               // ignore, don't care
            }
         }
      }
      
      return img == null ? null
                         : new ImageWrapper<BufferedImage>(img, img.getHeight(), img.getWidth());
   }
   
   /**
    * Get driver-specific child process factory.
    * 
    * @return   The reference to the factory.
    */
   @Override
   public ChildProcessFactory getChildProcessFactory()
   {
      return new ChildProcessFactory()
      {
         /**
          * Obtains a new instance of an interactive child process.
          *
          * @return   The new child process instance.
          */
         public InteractiveChildProcess getChildProcess()
         {
            // TODO: this requires a ChUI screen driver, to execute the emulator in it
            return null;
         }
      };
   }

   /**
    * Get BGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index.
    * 
    * @param    index
    *           Color number.
    *           
    * @return   instance of ColorRgb
    */
   @Override
   public ColorRgb getBgColor(int index)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;

      int windowId = ews.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      
      if (window != null)
      {
         rgb = ColorManager.getBgColor(window, index);
      }
      
      return rgb;
   }

   /**
    * Get FGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of ColorRgb
    */
   @Override
   public ColorRgb getFgColor(int index)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      
      int windowId = ews.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      
      if (window != null)
      {
         rgb = ColorManager.getFgColor(window, index);
      }
      
      return rgb;
   }
 
   /**
    * Get a system color.  Because system colors cannot be change at runtime a cache is used in
    * order to optimize the creation of <code>ColorRgb</code> instances.  
    * 
    * @param   index
    *          Color index.
    *          
    * @return  see above.
    */
   @Override
   public ColorRgb getSysColor(int index)
   {
      // first lookup in cache
      ColorRgb color = systemColorsCache.get(index);
      
      if (color == null)
      {
         // if not found in cache get from system colors table.
         color = ColorManager.getSysColor(index);
         
         // cache value
         systemColorsCache.put(index, color);
      }
      
      return color;
   }

   /**
    * Get foreground color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   @Override
   public ColorRgb getFgStdColor(String stdColor)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      
      int windowId = ews.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      
      if (window != null)
      {
         rgb = ColorManager.getStdColor(window, stdColor, true);
      }
      
      return rgb;
   }

   /**
    * Get background color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param    stdColor
    *           Color name.
    *           
    * @return   RGB value.
    */
   @Override
   public ColorRgb getBgStdColor(String stdColor)
   {
      ColorRgb rgb = ColorManager.DEFAULT_RGB;
      
      int windowId = ews.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      
      if (window != null)
      {
         rgb = ColorManager.getStdColor(window, stdColor, false);
      }
      
      return rgb;
   }
   
   /**
    * Translates the origin of the <code>Graphics2D</code> context to the
    * point (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
    * Modifies the <code>Graphics2D</code> context so that its new origin
    * corresponds to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * <code>Graphics2D</code> context's former coordinate system.  
    * All coordinates used in subsequent rendering operations on this graphics
    * context are relative to this new origin.
    *
    * @param    x 
    *           the specified x coordinate
    * @param    y 
    *           the specified y coordinate
    */
   @Override
   public void translate(int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.TRANSLATE);
      ps.x = x;
      ps.y = y;
      
      ews.offer(ps);

      // this is mutable...
      translatedOrigin.translate(x, y);
   }  
   
   /**
    * Draw a string at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   @Override
   public void drawString(String text, int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING);
      ps.text = text;
      ps.x = x;
      ps.y = y;
      
      ews.offer(ps);
   }      

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param    text
    *           text string to draw.
    * @param    legacyWidth
    *           The legacy width, which needs to be matched when drawing.
    * @param    legacyHeight
    *           The legacy width, which needs to be matched when drawing.
    * @param    x
    *           left position.
    * @param    y
    *           top position.
    */
   @Override
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING_SCALED);
      ps.text = text;
      ps.x = x;
      ps.y = y;
      ps.width = legacyWidth;
      ps.height = legacyHeight;
      
      ews.offer(ps);
   }

   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param    x1  
    *           the first point's <i>x</i> coordinate.
    * @param    y1  
    *           the first point's <i>y</i> coordinate.
    * @param    x2  
    *           the second point's <i>x</i> coordinate.
    * @param    y2  
    *           the second point's <i>y</i> coordinate.
    */
   @Override
   public void drawLine(int x1, int y1, int x2, int y2)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_LINE);
      ps.x = x1;
      ps.y = y1;
      ps.width = x2;
      ps.height = y2;
      
      ews.offer(ps);
   }      

   /**
    * Draw a rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    */   
   @Override
   public void drawRect(int x, int y, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      
      ews.offer(ps);
   }      

   /**
    * Draw a rounded rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    arcDiameter
    *           diameter of the arc at corners.
    */
   @Override
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_ROUND_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.arcDiameter = arcDiameter;
      
      ews.offer(ps);
   }      

   /**
    * Fill a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   @Override
   public void fillRect(int x, int y, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      
      ews.offer(ps);
   }      

   /**
    * Fill a rounded rectangle.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           rectangle width.
    * @param    height
    *           rectangle height.
    * @param    arcDiameter
    *           diameter of the arc at corners.
    */   
   @Override
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_ROUND_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.arcDiameter = arcDiameter;
      
      ews.offer(ps);
   }      

   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>ColorRgb</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param    x 
    *           the x coordinate of the rectangle to be drawn.
    * @param    y 
    *           the y coordinate of the rectangle to be drawn.
    * @param    width 
    *           the width of the rectangle to be drawn.
    * @param    height 
    *           the height of the rectangle to be drawn.
    * @param    raised 
    *           a boolean that determines whether the rectangle appears 
    *           to be raised above the surface or sunk into the surface.
    */
   @Override
   public void draw3DRect(int x, int y, int width, int height, boolean raised)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_3D_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.raised = raised;
      
      ews.offer(ps);
   }      

   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>ColorRgb</code>.  This method uses
    * the current <code>ColorRgb</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param    x 
    *           the x coordinate of the rectangle to be filled.
    * @param    y 
    *           the y coordinate of the rectangle to be filled.
    * @param    width
    *           the width of the rectangle to be filled.
    * @param    height 
    *           the height of the rectangle to be filled.
    * @param    raised 
    *           a boolean value that determines whether the rectangle appears 
    *           to be raised above the surface or etched into the surface.
    */   
   @Override
   public void fill3DRect(int x, int y, int width, int height, boolean raised)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_3D_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.raised = raised;
      
      ews.offer(ps);
   }      
   
   /**
    * Fill polygon.
    * 
    * @param    xPoints
    *           X coordinates.
    * @param    yPoints
    *           Y coordinates.
    */   
   @Override
   public void fillPolygon(int[] xPoints, int[] yPoints)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_POLYGON);
      ps.xPoints = xPoints;
      ps.yPoints = yPoints;
      ps.nPoints = Math.min(xPoints.length, yPoints.length);
      
      ews.offer(ps);
   }
   
   /**
    * Draw a image at the specified location.
    * 
    * @param    img
    *           Image to draw.
    * @param    x
    *           left location.
    * @param    y
    *           top location.
    * @param    w
    *           the image width.
    * @param    h
    *           the image height.
    * @param    xoff
    *           x offset inside image to start drawing.
    * @param    yoff
    *           y offset inside image to start drawing.
    * @param    transparent
    *           Flag indicating to use lower-left pixel as color marker to draw part of image as
    *           transparent.
    * @param    stretch
    *           Flag indicating stretch or shrink the image to fit display area.
    * @param    retain
    *           Flag indicating keep original image aspect ratio when stretching the image.
    * @param    convert3D
    *           Flag indicating autoconvert some image colors to the corresponding system wide 3D
    *           colors.
    */
   @Override
   public void drawImage(ImageWrapper img,
                         int          x,
                         int          y,
                         int          w,
                         int          h,
                         int          xoff,
                         int          yoff,
                         boolean      transparent,
                         boolean      stretch,
                         boolean      retain,
                         boolean      convert3D)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_IMAGE);
      ps.x = x;
      ps.y = y;
      ps.width = w;
      ps.height = h;
      ps.xOffset = xoff;
      ps.yOffset = yoff;
      ps.transparent = transparent;
      ps.stretchToFit = stretch;
      ps.retainShape = retain;
      ps.convert3D = convert3D;
      ps.img = img;

      ews.offer(ps);
   }

   /**
    * Draws a highlighted rectangular area of the specified color and bounds.
    * The color is alpha-blended with the content of the screen buffer.
    * 
    * @param    x
    *           left position
    * @param    y
    *           top position
    * @param    width
    *           area width
    * @param    height
    *           area height
    * @param    color
    *           The highlight color.
    */
   @Override
   public void drawHighlight(int x, int y, int width, int height, ColorRgb color)
   {
      if (width <= 0 || height <= 0)
      {
         return;
      }
      
      clearHighlight();
      
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_HIGHLIGHT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.color = color;
      
      ews.offer(ps);
   }

   /**
    * Clears the highlighted area drawn by the previous call of 
    * {@link #drawHighlight(int, int, int, int, ColorRgb)}.
    */
   @Override
   public void clearHighlight()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR_HIGHLIGHT);
      
      ews.offer(ps);
   }

   /**
   * Set current color used on draw operations.
   * 
   * @param    color
   *           Drawing color.
   */    
   @Override
   public void setColor(ColorRgb color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_COLOR);
      ps.color = color;

      ews.offer(ps);
   }
   
   /**
    * Set the window title at the OS level.
    * 
    * @param    title
    *           The window's title.
    */
   @Override
   public void setTitle(String title)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_TITLE);
      ps.title = title;

      ews.offer(ps);
   }
   
   /**
    * Set current font used on draw operations.
    * 
    * @param    font
    *           Drawing font.
    */
   @Override
   public void setGuiFont(FontDetails<F> font)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_FONT);
      ps.font = font;

      ews.offer(ps);
   }

   /**
    * Get the width of the specified font.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The font's width.
    */
   @Override
   public int getFontWidth(FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      int avgWidth = 0;
      int charno = 255;
      
      for (char ch = 1; ch <= 255; ch++)
      {
         avgWidth += fm.charWidth(ch);
      }
      avgWidth = (int) Math.ceil((avgWidth + 1.0) / (double) charno);

      return avgWidth;
   }

   /**
    * Get the maximum font width, for the given font.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The font's maximum width.
    */
   @Override
   public int getFontMaxWidth(FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      int[] widths = fm.getWidths();
      int maxWidth = 0;
      
      for (int w : widths)
      {
         maxWidth = Math.max(w,  maxWidth);
      }
      
      return maxWidth;
   }
   
   /**
    * Get the height of the specified font.
    * 
    * @param    fd
    *           The font details.
    *           
    * @return   The fonts's height.
    */
   @Override
   public int getFontHeight(FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      return fm.getHeight();
   }

   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   @Override
   public int getTextWidth(String text, FontDetails<F> fd)
   {
      FontMetricsHelper fm = ews.getFontMetrics(fd);
      
      return fm.stringWidth(text);
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    fd
    *           The font details used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, FontDetails<F> fd)
   {
      return ews.getTextHeight(text, fd);
   }
   
   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.
    * @param    legacyWidth
    *           The legacy width which needs to be matched.
    * @param    legacyHeight
    *           The legacy height which needs to be matched.
    *           
    * @return   The scaled font.
    */
   @Override
   public FontDetails<F> scaleFont(FontDetails<F> font, int legacyWidth, int legacyHeight)
   {
      FontMetricsHelper fm = ews.getFontMetrics(font);
   
      int avgWidth = 0;
      for (int ch = 0x20; ch <= 0xff; ch++)
      {
         avgWidth += fm.charWidth((char) ch);
      }
      avgWidth = (int) Math.ceil((double) avgWidth / (double) (0xff - 0x20 + 1));
      int height = fm.getHeight();
      
      FontDetails<F> scaled = font;
      
      if (height != legacyHeight || avgWidth != legacyWidth)
      {
         double scaleX = (double) legacyWidth / (double) avgWidth;
         double scaleY = (double) legacyHeight / (double) height;
         
         scaled = scaled.copy();
   
         deriveFont(scaled, scaleX, scaleY);
      }
      
      return scaled;
   }
   
   /**
    * Scale the given font by the specified X and Y factors.
    *
    * @param    fd
    *           Font to scale.  This instance's contained font will be modified. 
    * @param    scaleX
    *           X scaling factor.
    * @param    scaleY
    *           Y scaling factor.
    */
   public abstract void deriveFont(FontDetails<F> fd, double scaleX, double scaleY);
   
   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param    style
    *           Font style.
    */
   @Override
   public void setFontStyle(FontStyle style)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_FONT_STYLE);
      ps.style = style;
      
      ews.offer(ps);
   }
   
   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param    stroke
    *           Line style.
    */
   @Override
   public void setLineStroke(LineStroke stroke)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LINE_STROKE);
      ps.stroke = stroke;
      
      ews.offer(ps);
   }
   
   /**
    * The real location where the origin was {@link #translatedOrigin translated}.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint translatedOrigin()
   {
      return new NativePoint(translatedOrigin);
   }
   
   /**
    * Resize window.
    * 
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   @Override
   public void resizeWindow(int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESIZE_WINDOW);
      ps.width = width;
      ps.height = height;
      
      ews.offer(ps);
      ews.resizeWindow(width, height);
   }

   /**
    * Set window location.
    * 
    * @param    left
    *           left coordinate
    * @param    top
    *           top coordinate
    */
   @Override
   public void setWindowLocation(int left, int top)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LOCATION);
      ps.x = left;
      ps.y = top;
      
      ews.offer(ps);
      ews.repaint();
   }
   
   /**
    * Check if window has an emulator for the given window id.
    * 
    * @param    windowId
    *           Window ID.
    *           
    * @return   <code>true</code> if window has an associated emulator, <code>true</code>
    *           otherwise.
    */
   @Override
   public boolean hasWindowEmulator(int windowId)
   {
      return direct.hasWindowEmulator(windowId);
   }

   /**
    * Register a new window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void registerWindow(int windowId)
   {
      EmulatedWindowState ews = newEmulator();
      ews.captureMouseEvents(true);
      
      direct.registerWindow(windowId, ews);
   }

   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void deregisterWindow(int windowId)
   {
      direct.deregisterWindow(windowId);
      
      if (ews != null && windowId == ews.getWindowId())
      {
         ews = null;
      }
   }
   
   /**
    * Create a new window emulator.
    * 
    * @return   The newly instantiated window emulator.
    */
   public abstract EmulatedWindowState newEmulator();

   /**
    * On a GUI interface multiple windows could be created. Each window has it own renderer.
    * When a window is draw first the window must call this method to setup it renderer as 
    * the current graphical canvas.  
    * 
    * @param   windowId
    *          Window ID.
    */
   @Override
   public void selectWindow(int windowId)
   {
      if (ews == null || ews.getWindowId() != windowId)
      {
         direct.selectWindow(windowId);                           
         
         ews = direct.getWindowEmulator();
      }
   }   
   
   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of 
    *           this window.
    */
   @Override                                       
   public void iconifyWindow(int sourceId)
   {
      ews.iconifyWindow(sourceId);
   }
   
   /**
    * De-iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was deiconified, and triggered restoration of 
    *           this window.
    */
   @Override
   public void deiconifyWindow(int sourceId)
   {     
      ews.deiconifyWindow(sourceId);
   }

   /**
    * Check if the specified window is iconified or not.
    * 
    * @param     windowId
    *            The window ID.
    *            
    * @return    See above.
    */
   @Override
   public boolean isIconified(int windowId)
   {
      return direct.getWindowEmulator(windowId).isIconified();
   }
   
   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    * 
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void draw(NativePoint origin, NativeRectangle clip, Runnable core)
   {
      translate(origin.x, origin.y);

      if (clip != null)
      {
         setDrawingArea(clip);
      }
      
      try
      {
         core.run();
      }
      
      finally
      {
         if (clip != null)
         {
            setDrawingArea(null);
         }
         // restore context
         translate(-origin.x, -origin.y);
      }
   }

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the window's real (0,0) origin (top-left corner) and then translating to the
    * specified location, before executing the drawing.
    * <p>
    * The translated origin (as before this was called) needs to be restored, before this call
    * ends.
    * 
    * @param    origin
    *           The drawing origin (relative to window's real origin).
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void drawFromOrigin(NativePoint origin, NativeRectangle clip, Runnable core)
   {
      NativePoint save = new NativePoint(this.translatedOrigin);
      translate(-save.x, -save.y);
      
      try
      {
         draw(origin, clip, core);
      }
      
      finally
      {
         // move current translated back to screen (0,0)
         translate(-this.translatedOrigin.x, -this.translatedOrigin.y);
         
         // translate to the saved origin
         translate(save.x, save.y);
      }
   }
   
   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   @Override
   public void setDrawingArea(NativeRectangle bounds)
   {
      PaintStructure ps = new PaintStructure(bounds == null ? PaintPrimitives.NO_CLIP  
                                                            : PaintPrimitives.CLIP);
      if (bounds != null)
      {
         ps.x = bounds.left();
         ps.y = bounds.top();
         ps.width = bounds.right() - bounds.left() + 1;
         ps.height = bounds.bottom() - bounds.top() + 1;
      }
      
      ews.offer(ps);
   }

   /**
    * Start a batch drawing routine.
    */
   @Override
   public void startBatch()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.START_BATCH);
      ews.offer(ps);
   }

   /**
    * End a batch drawing routine.
    */
   @Override
   public void endBatch()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.END_BATCH);
      ews.offer(ps);
      
      ews.endBatch();
   }

   /**
    * Force screen buffer draw.
    */
   @Override
   public void flush()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FLUSH);
      
      ews.offer(ps);      
   }
      
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   @Override
   public void setWindowVisible(boolean visible)
   {
      ews.setWindowVisible(visible);
   }
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      return ews.getWindowLocation();
   }

   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWindowDimension()
   {
      return new NativeDimension(ews.getWidth(), ews.getHeight());
   }
   
   /**
    * Force the GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   @Override
   public void stackWindows(int[] windowIds)
   {
      EmulatedWindowState oldEws = (ews == null ? null : ews);
      
      for (int wid : windowIds)
      {
         selectWindow(wid);
         ews.saveFocusListeners();
      }

      for (int wid : windowIds)
      {
         selectWindow(wid);
         ews.moveToTop();
      }

      for (int wid : windowIds)
      {
         selectWindow(wid);
         ews.restoreFocusListeners();
      }
      
      if (oldEws != null)
      {
         selectWindow(oldEws.getWindowId());
      }
   }

   /**
    * Determine if the given window is on top of all windows, including OS.
    * 
    * @param    windowId
    *           The window ID to check.
    *           
    * @return   See above.
    */
   @Override
   public boolean isOnTop(int windowId)
   {
      boolean top = false;
      
      try
      {
         // an exception will be thrown if window id is invalid
         EmulatedWindowState ews = direct.getWindowEmulator(windowId);
         
         top = ews.isOnTop();
      }
      
      catch (RuntimeException re)
      {
         // do nothing, eat it
      }
      
      return top;
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(int[] windowIds, boolean capture)
   {
      EmulatedWindowState oldEws = (ews == null ? null : ews);
      
      // TODO: do we really need to select each window or can we just use
      //       direct.getWindowEmulator() to get each one?  
      for (int wid : windowIds)
      {
         selectWindow(wid);

         ews.enableEvents(capture);
      }
      
      if (oldEws != null)
      {
         selectWindow(oldEws.getWindowId());
      }
   }
   
   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type.
    */
   @Override
   public void setCursor(CursorType type)
   {
      ews.setCursor(type);
   }

   /**
    * Set cursor status (ON/OFF).
    * <p>
    * For GUI, this does nothing.                  
    * 
    * @param   on
    *          <code>true</code> - show cursor, <code>false</code> - hide cursor.
    *
    * @return  Always <code>false</code>.
    */
   @Override
   public boolean setCursorStatus(boolean on)
   {
      return false;
   }

   /**
    * Reset current screen mode.
    * <p>
    * Upon exit from this method P2J assumes that screen is in correct state
    * and can be used for interaction with the user.
    */
   @Override
   public void resetMode()
   {
      // NOP for now, what should this do?
   }                                                              

    /**
     * Resume normal driver operation. If implementation is terminal-based,
     * they terminal should be acquired again and driver should be ready to 
     * perform regular screen I/O operations.
     */
   @Override
   public void resume()
   {
      // NOP for now, what should this do?
   }

   /**
    * Suspend all screen I/O operations. If implementation is terminal-based, 
    * then terminal should be released to allow other application executed 
    * from P2J access terminal.
    */
   @Override
   public void suspend()
   {
      // NOP for now, what should this do?
   }
   
   /**
    * Perform driver shutdown.
    */
   @Override
   public void shutdown()
   {
      // destroy all emulators
      direct.shutdown();
   }
   
   /**
    * Get the window UI theme name.  Can be either xp or classic.
    * 
    * @return   The currently used theme name for OS UI interface elemens.
    */
   @Override
   public String getUiTheme()
   {                                                                   
      // first call, request server for current theme to use
      if (theme == null)
      {
         Directory dir = DirectoryManager.getInstance();
         
         theme = dir.getString(Directory.ID_RELATIVE, "windows-os-ui-theme", "classic")
                    .toLowerCase();
      }
      
      return theme;
   }

   /** 
    * Get the native screen size, in pixel units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getScreenSize()
   {
      return ews.getScreenSize();
   }
   
   /**
    * Returns display size.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The dimensions in pixels.
    */
   @Override
   public NativeDimension getDisplaySize(int windowId)
   {
      // this will never be null because an exception will be thrown if window id is invalid
      EmulatedWindowState ews = direct.getWindowEmulator(windowId);
      
      return ews.getDisplaySize();
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea(int windowId)
   {
      // this will never be null because an exception will be thrown if window id is invalid
      EmulatedWindowState ews = direct.getWindowEmulator(windowId);
      
      return ews.getDisplayWorkArea();
   }

   /**
    * Check if the currently selected window is visible.
    * 
    * @return   See above.
    */
   @Override
   public boolean isWindowVisible()
   {
      return ews.isWindowVisible();
   }

   /**
    * Set status text message of current window.
    * 
    * @param    message
    *           Text message.
    */
   @Override
   public void setStatusText(String message)
   {
      Window<?> window = (Window<?>) WindowManager.findWindow(ews.getWindowId());
      setStatusText(window, message);
   }

   /**
    * Set status text message of specified window.
    * 
    * @param    message
    *           Text message.
    * @param    window
    *           Window instance.
    */
   @Override
   public void setStatusText(Window<?> window, String message)
   {
      if (window != null)
      {
         window.getStatusArea().setText(message);
      }      
   }
   
   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      return ews.captureMouseEvents(state);
   }

   /**
    * Handle the specified event, in the context of the specified widget.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   @Override
   public void handleMouseEvent(int sourceID, MouseEvent evt)
   {
      MouseHandler.handleMouseEvent(sourceID, evt);
   }
      
   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    wid
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   @Override
   public void listenAnyMouseEvent(int wid, long when)
   {
      MouseHandler.listenAnyMouseEvent(wid, when);
   }
   
   /**
    * Register the specified widget as a moveable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerMoveableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new MouseMoveable(wid));
   }

   /**
    * Register the specified widget as a resizeable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerResizeableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new MouseResizeable(wid));
   }

   /**
    * Register the specified widget for popup events, via mouse.
    * <p>
    * The popup with the given ID will be shown when the specified mouse button is pressed inside
    * the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   @Override
   public void registerWidgetPopup(int wid, int btn, int pid)
   {
      MouseHandler.registerWidgetAction(wid, new MousePopupable(wid, btn, pid));
   }

   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * 
    * @param    wid
    *           The widget id.
    */
   @Override
   public void registerEditableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new EditableWidget(wid));
   }

   /**
    * Register this widget as having tooltip functionality reacting on mouse entering and exiting
    * events.  The registration is pending if tooltip engine is not available for session or
    * tooltip text is empty.
    * 
    * @param    widget
    *           The widget object to register.
    * @param    fnt
    *           The font to draw tooltip text.
    * @param    now
    *           The flag indicating whether the action be registered immediately or later.
    */
   @Override
   public void registerTooltipWorker(Widget<GuiOutputManager> widget,
                                     FontDetails<F>           fnt,
                                     boolean                  now)
   {
      int wid = widget.getId().asInt();
      MouseTooltipWorker worker = new MouseTooltipWorker(wid, fnt);
      
      // now or delayed?
      if (now)
      {
         MouseHandler.registerWidgetAction(wid, worker);
      }
      else
      {
         pendingTtWorkers.put(widget, worker);
      }
   }

   /**
    * Register the tooltip workers if this process was dalayed with regular registration.
    * 
    * @param    wid
    *           The widget id. 0 means register all pending workers. 
    */
   @Override
   public void registerPendingTooltipWorkers(int wid)
   {
      // if tooltip feature is turned off for session nothing to do
      if (!ToolTip.getSessionTooltips())
      {
         return;
      }
      
      // handle pending workers
      if (wid == 0)
      {
         // all workers mode enumerate and register all valid
         Iterator<Widget> iter = pendingTtWorkers.keySet().iterator();
         while (iter.hasNext())
         {
            // get the widget object
            Widget w = iter.next();
            // handle single worker
            if (registerPendingTooltipWorkerInt(w))
            {
               iter.remove();
            }
         }
      }
      else
      {
         // handle single worker
         Widget<GuiOutputManager> w = (Widget<GuiOutputManager>) OutputManager.resolveWidget(wid);
         if (registerPendingTooltipWorkerInt(w))
         {
            pendingTtWorkers.remove(w);
         }
      }
   }
   
   /**
    * Deregister the tooltip workers already registered with mouse handler.
    * 
    * @param    wid
    *           The widget id. 0 means deregister all registered workers. 
    */
   @Override
   public void deregisterTooltipWorkers(int wid)
   {
      if (wid == 0)
      {
         // deregister all widgets mode
         Integer[] widList = MouseHandler.getWidgetListForTooltipWorkers();
         // enumerate all widget ID's if there is
         if (widList != null)
         {
            for (int i = 0; i < widList.length; i++)
            {
               deregisterTooltipWorkerInt(widList[i].intValue());
            }
         }
      }
      else
      {
         deregisterTooltipWorkerInt(wid);
      }
   }
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void deregisterWidget(int wid)
   {
      MouseHandler.deregisterWidget(wid);
      // clean the pending worker too but for real widgets only
      if (wid > 0)
      {
         pendingTtWorkers.remove((Widget<GuiOutputManager>) OutputManager.resolveWidget(wid));
      }
   }
   
   /**
    * Register the tooltip worker if this process was dalayed with regular registration.  Internal
    * version.
    * 
    * @param    widget
    *           The widget object. Null means register all pending workers. 
    *           
    * @return   <code>TRUE</code> if worker registered, <code>FALSE</code> otherwise.
    */
   private boolean registerPendingTooltipWorkerInt(Widget<GuiOutputManager> widget)
   {
      // simple validity check
      if (widget == null)
      {
         // not valid, exit
         return false;
      }
      // verify the text
      String text = UiUtils.getConfig(widget, BaseConfig.class).tooltip;
      if (text != null && text.length() > 0)
      {
         // remove worker from list of pendings
         MouseWidgetAction worker = pendingTtWorkers.get(widget);
         if (worker != null)
         {
            MouseHandler.registerWidgetAction(widget.getId().asInt(), worker);
            return true;
         }
      }
      // worker was not registered
      return false;
   }
   
   /**
    * Deregister the tooltip workers already registered with mouse handler.  Internal wersion.
    * 
    * @param    wid
    *           The widget id.
    */
   private void deregisterTooltipWorkerInt(int wid)
   {
      // extract handler for this widget
      MouseWidgetAction worker = MouseHandler.deregisterTooltipWorker(wid);

      // determine the widget object for this widget ID
      Widget<GuiOutputManager> widget =
         (Widget<GuiOutputManager>) OutputManager.resolveWidget(wid);
         
      // if widget or handler are null we do nothing 
      if (widget == null || worker == null)
      {
         // something is wrong, exit
         return;
      }
      
      // save the deregistered worker in pending map to be registered later 
      pendingTtWorkers.put(widget, worker);
   }
}   
