/*
** Module   : EmulatedWindowState.java
** Abstract : generic code for the core GUI window emulator
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20150429 First version with most functionality moved from GuiSimulator. There is a
**                  one-to-one relationship between instances of this class and a 4GL window.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.WindowEvent;

/**
 * This is generic code for the core GUI window emulator. A pool of window-specific drawing
 * operations is used to queue drawing primitives. The operations are enqueued in the pool
 * one at a time and then are drawn into a buffered image which is provided by the sub-class.
 * When the pool is empty, the screen is updated with the image from the stored buffered image.
 * <p>
 * There should not be any Swing/AWT-specific or Web-specific code in this class.
 * <p>
 * The actual GUI processing does not heavily utilize the pluggable <code>OutputPrimitives</code>
 * interface.  In other words, the interactive UI processing GUI is not switched in and out like
 * in ChUI.  For this reason, the core output processing in ChUI is driven using the primitives
 * approach, but in GUI we simply honor the minimum implementation in this class.  The majority
 * of this class is really about providing access to and management of the core window structure.
 * The primitives implementation is secondary, but the functionality is best implemented here.
 */
public abstract class EmulatedWindowState<F, I>
{
   /** The window ID associated with this emulated window. */
   protected int windowId;
   
   /** The registered mouse handler. */
   protected MouseHandler mouseHandler;
   
   /** Track the window's iconified state (which can be set at the OS or P2J level). */
   private boolean iconified = false;
   
   /** Draw operations pools */
   private Deque<PaintStructure<F, I>> graphicPool = new ConcurrentLinkedDeque<>();

   /** The bitmap associated with this simulator. */
   private ScreenBitmap bitmap;
   
   /**
    * Constructor.
    */
   public EmulatedWindowState()
   {
      mouseHandler = new MouseHandler(this);
   }

   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   public boolean captureMouseEvents(boolean state)
   {
      return mouseHandler.captureMouseEvents(state);
   }
   
   /**
    * Get the screen width, in character units.
    * 
    * @return    See above.
    */
   public int screenWidth()
   {
      return bitmap.cols();
   }

   /**
    * Get the screen height, in character units.
    * 
    * @return    See above.
    */
   public int screenHeight()
   {
      return bitmap.rows();
   }
   
   /**
    * Set the active screen bitmap for the current window.
    * 
    * @param    bitmap
    *           The screen bitmap.
    */
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      ScreenBitmap prev = bitmap;

      this.bitmap = bitmap;
      
      if (prev == null || bitmap.cols() != prev.cols() || bitmap.rows() != prev.rows())
      {
         resizeWindow(bitmap.cols(), bitmap.rows());
      }
   }

   /**
    * Get the screen bitmap for the current window.
    * 
    * @return   The current window's screen bitmap.
    */
   public ScreenBitmap getScreenBitmap()
   {
      return bitmap;
   }

   /**
    * Set the window ID for this simulator.
    * 
    * @param   windowId
    *          The window ID.
    */
   public void setWindowId(int windowId)
   {
      this.windowId = windowId;
   }

   /**
    * Get the window ID for this simulator.
    */
   public int getWindowId()
   {
      return windowId;
   }

   /**
    * Add a new <code>PaintStructure</code> to the drawing pool.
    * The item is enqueued at the end of the pool.
    * 
    * @param   ps
    *          A new PaintStructure instance.          
    */      
   public void offer(PaintStructure<F, I> ps)
   {
      // to prevent deadlocks, synchronize on the frame's layout lock
      synchronized (lock())
      {
         graphicPool.offerLast(ps);

         lock().notifyAll();
      }
   }

   /**
    * Resize window.
    * 
    * @param    width
    *           new width.
    * @param    height
    *           new height
    */
   public void resizeWindow(int width, int height)
   {
      repaint();

      this.bitmap.setRows(height);
      this.bitmap.setColumns(width);
   }
   
   /**
    * Manage any window state related to iconification.
    * 
    * @param    iconify
    *           <code>true</code> if this window is being iconified, <code>false</code> for
    *           de-iconified.
    */
   public void setIconificationState(boolean iconify)
   {
      this.iconified = iconify;
   }

   /**
    * Iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was iconified, and triggered iconification of 
    *           this window.
    */
   public void iconifyWindow(int sourceId)
   {
      if (iconified)
      {
         return;
      }
      
      setIconificationState(true);
      
      // raise the event here
      WindowGuiImpl window = (WindowGuiImpl) ThinClient.getInstance().getWidget(windowId);
      WindowEvent evt = new WindowEvent(window, Keyboard.SE_WINDOW_MINIMIZED);
      evt.setOrigin(sourceId);
      ThinClient.getInstance().postOSEvent(evt);
   }

   /**
    * De-iconify the window.
    * 
    * @param    sourceId
    *           The root ID of the sub-tree which was deiconified, and triggered restoration of 
    *           this window.
    */
   public void deiconifyWindow(int sourceId)
   {
      if (!iconified)
      {
         return;
      }
      
      setIconificationState(false);
      
      // some window systems don't always generate a window deiconified event, so post one
      // manually... setting the "iconified" flag prevents a double-event generation, in case one
      // is raised by the window system too
      WindowGuiImpl window = (WindowGuiImpl) ThinClient.getInstance().getWidget(windowId);
      WindowEvent evt = new WindowEvent(window, Keyboard.SE_WINDOW_RESTORED);
      evt.setOrigin(sourceId);
      ThinClient.getInstance().postOSEvent(evt);
   }
   
   /**
    * Check if this window is iconified or not.
    * 
    * @return    See above.
    */
   public boolean isIconified()
   {
      return iconified;
   }
   
   /**
    * End a batch drawing routine.
    */
   public void endBatch()
   {
      repaint();
      
      synchronized (lock())
      {
         lock().notifyAll();
      }
   }
   
   /**
    * Save and remove the OS window's focus listeners.
    */
   public abstract void saveFocusListeners();

   /**
    * Restore the previously saved OS window's focus listeners.
    */
   public abstract void restoreFocusListeners();

   /**
    * Move this window to top, at the OS level.
    */
   public abstract void moveToTop();
   
   /**
    * Determine if the window is on top of all windows, including OS.
    * 
    * @return   <code>true</code> if the window is topmost.
    */
   public abstract boolean isOnTop();

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   public abstract void enableEvents(boolean capture);
   
   /** 
    * Get the native screen size, in pixel units.
    * 
    * @return   The native screen size.
    */
   public abstract NativeDimension getScreenSize();
   
   /**
    * Returns display size.
    * 
    * @return   The dimensions in pixels.
    */
   public abstract NativeDimension getDisplaySize();

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * @return   The area in pixels.
    */
   public abstract NativeRectangle getDisplayWorkArea();
   
   /**
    * Get the window location on screen.
    * 
    * @return   This window's location.
    */
   public abstract NativePoint getWindowLocation();
   
   /**
    * Get the metrics of the given font.
    *  
    * @param    font
    *           The font for which the font metrics are to be obtained.
    *           
    * @return   The font metrics.
    */
   public abstract FontMetricsHelper getFontMetrics(FontDetails<F> font);

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   public abstract int getTextHeight(String text, FontDetails<F> font);
   
   /**
    * Get the window height.
    * 
    * @return   This window's width in pixels.
    */
   public abstract int getHeight();
   
   /**
    * Get the window width.
    * 
    * @return   This window's width in pixels.
    */
   public abstract int getWidth();
   
   /**
    * Render the given paint operation.
    * 
    * @param    ps
    *           Paint operation details.
    */
   public abstract void draw(PaintStructure<F, I> ps);
   
   /**
    * Cause repaint processing to occur.
    */
   public abstract void repaint();
   
   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type.
    */
   public abstract void setCursor(CursorType type);
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public abstract void setWindowVisible(boolean visible);
   
   /**
    * Check if the simulated window is visible.
    * 
    * @return   See above.
    */
   public abstract boolean isWindowVisible();
   
   /**
    * Render the given paint operation.  Any graphics context or other resources needed must
    * have already been setup/stored by the caller.
    */
   public abstract void flush();
   
   /**
    * Release resources as the window is being terminated.
    */
   public abstract void quit();
   
   /**
    * Obtain the lock object that can be used to synchronize drawing operations.
    *
    * @return   The shared lock object.
    */
   protected abstract Object lock();

   /**
    * Update screen by iterating over the pool and render each pending operation. Any graphics
    * context or other resources needed must have already been setup/stored by the caller.
    */
   protected void drawOnScreenBitmap()
   {
      // ensure the drawing will be performed by only one thread at a time
      synchronized (lock())
      {
         int waitForBatch = 0;
         
         while (!graphicPool.isEmpty() || waitForBatch != 0)
         {
            // if a batch was started, wait for it to end.
            if (graphicPool.isEmpty() && waitForBatch != 0)
            {
               try
               {
                  lock().wait();
               }
               catch (InterruptedException e)
               {
                  // terminate painting
                  break;
               }
            }
            
            PaintStructure<F, I> ps = graphicPool.pollFirst();

            if (ps != null)
            {
               switch (ps.id)
               {
                  case FLUSH:
                     flush();
                     break;
                  
                  case START_BATCH:
                     waitForBatch++;
                     break;

                  case END_BATCH:
                     waitForBatch--;
                     break;
                     
                  default:
                     draw(ps);
                     break;
               }
            }
         }
   
         flush();
      }
   }
}
