/*
** Module   : GuiPrimitives.java
** Abstract : GUI version of the output primitives
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Added mouse event support.  More misc changes related to window behavior.
** 004 MAG 20140905 Added fillPoly method.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Changed drawImage, to specify the image's width and height.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 EVL 20141116 Adding round rectangles draw and fill.
** 011 EVL 20141125 Adding dotted line draw capability.
** 012 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 013 HC  20141207 Implemented runtime support for several SESSION attributes.
** 014 EVL 20150203 Adding image offset drawing capability.
** 015 EVL 20150218 Adding support TRANSPARENT, STRETCH-TO-FIT, RETAIN-SHAPE and CONVERT-3D-COLORS
**                  options to image drawing.
** 016 CA  20150317 Added API to determine the font's max width.
** 017 CA  20150408 Added APIs to set the OS window's title and to de-iconify the window.
** 018 GES 20150501 Minimized the APIs needed here to only those relaced to managing or accessing
**                  the registered windows and/or the current window. This class is now a kind of
**                  router that keeps the list of emulated window instances and a current
**                  instance which can be swapped.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import com.goldencode.p2j.ui.client.*;

/** 
 * This is a low level GUI-specific window emulator management interface.  Real drawing does not
 * occur in this class. This is a kind of router that keeps the list of emulated window instances
 * and a current instance which can be swapped.
 * <p>
 * This interface is for GUI-specific features. It is intended to be specific to a given
 * UI-modality (GUI in this case). No driver-specific implementation classes or features
 * should be present.  For example, no Swing-specific feature usage is allowed.
 */
public interface GuiPrimitives
extends OutputPrimitives
{   
   /**
    * Get the currently selected window emulator.
    * 
    * @return    The current window emulator.
    */
   public EmulatedWindowState getWindowEmulator();
   
   /**
    * Returns the window emulator registered for the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *           
    * @return   The identified window emulator.
    * 
    * @throws   RuntimeException
    *           When no window emulator is registered for the supplied window id.
    */
   public EmulatedWindowState getWindowEmulator(int windowId);
   
   /**
    * Check if window has an emulator for the given window id.
    * 
    * @param    windowId
    *           Window ID.
    *           
    * @return   <code>true</code> if window has an associated emulator, <code>true</code>
    *           otherwise.
    */
   public boolean hasWindowEmulator(int windowId);
   
   /**
    * Register a new window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    ews
    *           New emulated window state to register.  
    */
   public void registerWindow(int windowId, EmulatedWindowState ews);

   /**
    * Deregister the given window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   public void deregisterWindow(int windowId);
   
   /**
    * On a GUI interface multiple windows can be created. Each window has it own emulator.
    * When a window is first drawn, this method is used to set it's emulator as the current 
    * target for output operations.  
    * 
    * @param    windowId
    *           Window ID.
    */
   public void selectWindow(int windowId);
   
   /**
    * Perform shutdown.
    */
   public void shutdown();
}
