/*
** Module   : MouseTooltipWorker.java
** Abstract : Implementation of a mouse action which shows a tooltip.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 EVL 20141211 Created initial version.
** 002 GES 20150501 Moved to a more generic package.
*/

package com.goldencode.p2j.ui.client.gui.driver;

// this usage of AWT events is intentional, they are being used separately from the AWT/Swing
// event model; no other AWT/Swing usage is occurring in this class
import java.awt.event.*;

import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;

/**
 * Listener for mouse actions which can trigger a tooltip to be shown for a specified widget.
 */
public class MouseTooltipWorker
extends MouseWidgetAction
{
   /** The widget's tooltip renderer. */
   private ToolTip toolTip = null;
   
   /** Special thread for tooltip life cycle. */
   private Thread ttThread = null;
   
   /** Font to draw the tooltip. */
   private FontDetails<?> fnt = null;
   
   /**
    * Create a new mouse tooltip worker, for the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    ttFnt
    *           The font to draw the tooltip with given widget.
    */
   public MouseTooltipWorker(int wid, FontDetails<?> ttFnt)
   {
      super(wid);
      
      this.fnt = ttFnt;
   }

   /**
    * Notification of a mouse entered event occurred for this widget.  Start point to draw
    * tooltip.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      // this is for tooltip usage
      if (widget.isEnabled())
      {
         // first time tooltip usage, we need to create it first
         if (toolTip == null)
         {
            toolTip = new ToolTip(widget, fnt);
         }
         
         // and show
         toolTip.setOrigin(new NativePoint(e.getX(), e.getY()));
         
         // execute in separate thred
         ttThread = new Thread(toolTip);
         ttThread.start();
      }
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  Disappear tooltip.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      // this is for tooltip usage
      if (widget.isEnabled() && toolTip != null)
      {
         // tooltip should disappear if alive
         if (ttThread.isAlive())
         {
            ttThread.interrupt();
         }
      }
   }
}
